import numpy as np

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.saes import SAES


class SAMAES(SAES):
    """Self-Adaptation Matrix Adaptation Evolution Strategy (SAMAES).

    .. note:: It is **highly recommended** to first attempt more advanced ES variants (e.g. `LMCMA`, `LMMAES`) for
       large-scale black-box optimization. Here we include it only for *benchmarking* and *theoretical* purpose.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`),
                * 'lr_sigma'      - learning rate of global step-size adaptation (`float`, default:
                  `1.0/np.sqrt(2*problem['ndim_problem'])`).
                * 'lr_matrix'     - learning rate of matrix adaptation (`float`, default:
                  `1.0/(2.0 + ((problem['ndim_problem'] + 1.0)*problem['ndim_problem'])/options['n_parents'])`).

    Examples
    --------
    Use the optimizer `SAMAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.samaes import SAMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> samaes = SAMAES(problem, options)  # initialize the optimizer class
       >>> results = samaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SAMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SAMAES: 5000, 5.674972930091687e-17

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/56k42a2n>`_ for more details.

    Attributes
    ----------
    lr_sigma      : `float`
                    learning rate of global step-size adaptation.
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    sigma         : `float`
                    final global step-size, aka mutation strength.
    lr_matrix     : `float`
                    learning rate of matrix adaptation.

    References
    ----------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700). ACM.
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870
    """
    def __init__(self, problem, options):
        SAES.__init__(self, problem, options)
        if self.lr_sigma is None:
            self.lr_sigma = 1.0/np.sqrt(2.0*self.ndim_problem)
        self.lr_matrix = 1.0/(2.0 + ((self.ndim_problem + 1.0)*self.ndim_problem)/self.n_parents)
        self._eye = np.eye(self.ndim_problem)  # for matrix adaptation

    def initialize(self, is_restart=False):
        x, mean, sigmas, y = SAES.initialize(self, is_restart)
        m = np.eye(self.ndim_problem)  # for matrix adaptation
        return x, mean, sigmas, y, m

    def iterate(self, x=None, mean=None, sigmas=None, y=None, m=None, args=None):
        z = np.empty((self.n_individuals, self.ndim_problem))
        d = np.empty((self.n_individuals, self.ndim_problem))
        for k in range(self.n_individuals):  # to sample offspring population
            if self._check_terminations():
                return x, sigmas, y, m, z, d
            sigmas[k] = self.sigma*np.exp(self.lr_sigma*self.rng_optimization.standard_normal())
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            d[k] = np.matmul(m, z[k])
            x[k] = mean + sigmas[k]*d[k]
            y[k] = self._evaluate_fitness(x[k], args)
        return x, sigmas, y, m, z, d

    def restart_initialize(self, x=None, mean=None, sigmas=None, y=None, m=None):
        if self.is_restart and self._restart_initialize(y):
            x, mean, sigmas, y, m = self.initialize(True)
        return x, mean, sigmas, y, m

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, sigmas, y, m = self.initialize()
        while True:
            # sample and evaluate offspring population
            x, sigmas, y, m, z, d = self.iterate(x, mean, sigmas, y, m, args)
            if self._check_terminations():
                break
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            order = np.argsort(y)[:self.n_parents]
            mean = np.mean(x[order], axis=0)  # intermediate multi-recombination
            self.sigma = np.mean(sigmas[order])  # intermediate multi-recombination
            # use the following code (fast version) owing to its quadratic time complexity
            dz = np.zeros((self.ndim_problem, self.ndim_problem))  # for matrix adaptation
            for i in range(self.n_parents):
                dz += np.outer(d[order[i]], z[order[i]])
            m = (1.0 - self.lr_matrix)*m + self.lr_matrix*(dz/self.n_parents)
            x, mean, sigmas, y, m = self.restart_initialize(x, mean, sigmas, y, m)
        return self._collect(fitness, y, mean)



"""A simple example for `COCO` Benchmarking using `PyPop7`:
    https://github.com/numbbo/coco
    
    To install `COCO` successfully, please read the above link carefully.
"""
import os
import webbrowser  # for post-processing in the browser

import numpy as np
import cocoex  # experimentation module of `COCO`
import cocopp  # post-processing module of `COCO`

from pypop7.optimizers.es.maes import MAES


if __name__ == '__main__':
    suite, output = 'bbob', 'coco-maes'
    budget_multiplier = 1e3  # or 1e4, 1e5, ...
    observer = cocoex.Observer(suite, 'result_folder: ' + output)
    minimal_print = cocoex.utilities.MiniPrint()
    for function in cocoex.Suite(suite, '', ''):
        function.observe_with(observer)  # generate data for `cocopp` post-processing
        sigma = np.min(function.upper_bounds - function.lower_bounds)/3.0
        problem = {'fitness_function': function,
                   'ndim_problem': function.dimension,
                   'lower_boundary': function.lower_bounds,
                   'upper_boundary': function.upper_bounds}
        options = {'max_function_evaluations': function.dimension*budget_multiplier,
                   'seed_rng': 2022,
                   'x': function.initial_solution,
                   'sigma': sigma}
        solver = MAES(problem, options)
        print(solver.optimize())
    cocopp.main(observer.result_folder)
    webbrowser.open('file://' + os.getcwd() + '/ppdata/index.html')



import numpy as np  # engine for numerical computing

from pypop7.optimizers.nes.ones import ONES


def _combine_block(ll):
    ll = [list(map(np.mat, row)) for row in ll]
    h = [m.shape[1] for m in ll[0]]
    v, v_i = [row[0].shape[0] for row in ll], 0
    mat = np.zeros((sum(h), sum(v)))
    for i, row in enumerate(ll):
        h_i = 0
        for j, m in enumerate(row):
            mat[v_i:v_i + v[i], h_i:h_i + h[j]] = m
            h_i += h[j]
        v_i += v[i]
    return mat


class ENES(ONES):
    """Exact Natural Evolution Strategy (ENES).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring/descendants, aka offspring population size (`int`),
                * 'n_parents'     - number of parents/ancestors, aka parental population size (`int`),
                * 'mean'          - initial (starting) point (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'lr_mean'       - learning rate of distribution mean update (`float`, default: `1.0`),
                * 'lr_sigma'      - learning rate of global step-size adaptation (`float`, default: `1.0`).

    Examples
    --------
    Use the optimizer `ENES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy  # engine for numerical computing
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.nes.enes import ENES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> enes = ENES(problem, options)  # initialize the optimizer class
       >>> results = enes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"ENES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       ENES: 5000, 0.00035668252927080496

    Attributes
    ----------
    lr_mean       : `float`
                    learning rate of distribution mean update.
    lr_sigma      : `float`
                    learning rate of global step-size adaptation.
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search/sampling/mutation distribution.
    n_individuals : `int`
                    number of offspring/descendants, aka offspring population size.
    n_parents     : `int`
                    number of parents/ancestors, aka parental population size.
    sigma         : `float`
                    global step-size, aka mutation strength (i.e., overall std of Gaussian search distribution).

    References
    ----------
    Wierstra, D., Schaul, T., Glasmachers, T., Sun, Y., Peters, J. and Schmidhuber, J., 2014.
    Natural evolution strategies.
    Journal of Machine Learning Research, 15(1), pp.949-980.
    https://jmlr.org/papers/v15/wierstra14a.html

    Schaul, T., 2011.
    Studies in continuous black-box optimization.
    Doctoral Dissertation, Technische Universität München.
    https://people.idsia.ch/~schaul/publications/thesis.pdf

    Yi, S., Wierstra, D., Schaul, T. and Schmidhuber, J., 2009, June.
    Stochastic search using the natural gradient.
    In International Conference on Machine Learning (pp. 1161-1168). ACM.
    https://dl.acm.org/doi/abs/10.1145/1553374.1553522

    See the official Python source code from PyBrain:
    https://github.com/pybrain/pybrain/blob/master/pybrain/optimization/distributionbased/nes.py
    """
    def __init__(self, problem, options):
        ONES.__init__(self, problem, options)
        if options.get('lr_mean') is None:
            self.lr_mean = 1.0
        if options.get('lr_sigma') is None:
            self.lr_sigma = 1.0

    def _update_distribution(self, x=None, y=None, mean=None, cv=None):
        order = np.argsort(-y)
        u = np.empty((self.n_individuals,))
        for i, o in enumerate(order):
            u[o] = self._u[i]
        inv_d, inv_cv = np.linalg.inv(self._d_cv), np.linalg.inv(cv)
        dd = np.diag(np.diag(inv_d))
        v = np.zeros((self._n_distribution, self.n_individuals))
        for k in range(self.n_individuals):
            diff = x[k] - mean
            s = np.dot(inv_d.T, diff)
            v[:self.ndim_problem, k] += diff
            v[self.ndim_problem:, k] += self._triu2flat(np.outer(s, np.dot(inv_d, s)) - dd)
        j = self._n_distribution - 1
        g = 1.0/(inv_cv[-1, -1] + np.square(inv_d[-1, -1]))
        d = 1.0/inv_cv[-1, -1]
        v[j, :] = np.dot(g, v[j, :])
        j -= 1
        for k in reversed(list(range(self.ndim_problem - 1))):
            w = inv_cv[k, k]
            w_g = w + np.square(inv_d[k, k])
            q = np.dot(d, inv_cv[k + 1:, k])
            c = np.dot(inv_cv[k + 1:, k], q)
            r, r_g = 1.0/(w - c), 1.0/(w_g - c)
            t, t_g = -(1.0 + r*c)/w, -(1.0 + r_g*c)/w_g
            g = _combine_block([[r_g, t_g*q], [np.mat(t_g*q).T, d + r_g*np.outer(q, q)]])
            d = _combine_block([[r, t*q], [np.mat(t*q).T, d + r*np.outer(q, q)]])
            v[j - (self.ndim_problem - k - 1):j + 1, :] = np.dot(g, v[j - (self.ndim_problem - k - 1):j + 1, :])
            j -= self.ndim_problem - k
        grad, v_2 = np.zeros((self._n_distribution,)), v*v
        j = self._n_distribution - 1
        for k in reversed(list(range(self.ndim_problem))):
            base = np.sum(v_2[j - (self.ndim_problem - k - 1):j + 1, :], 0)
            grad[j - (self.ndim_problem - k - 1):j + 1] = np.dot(
                v[j - (self.ndim_problem - k - 1):j + 1, :], u - np.dot(base, u)/np.sum(base))
            j -= self.ndim_problem - k
        base = np.sum(v_2[:j + 1, :], 0)
        grad[:j + 1] = np.dot(v[:j + 1, :], (u - np.dot(base, u)/np.sum(base)))
        grad /= self.n_individuals
        mean += self.lr_mean*grad[:self.ndim_problem]
        self._d_cv += self.lr_sigma*self._flat2triu(grad[self.ndim_problem:])
        cv = np.dot(self._d_cv.T, self._d_cv)
        return x, y, mean, cv



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class CC(Optimizer):
    """Cooperative Coevolution (CC).

    This is the **abstract** class for all `CC` classes. Please use any of its instantiated subclasses to
    optimize the black-box problem at hand.

    .. note:: `CC` uses the **decomposition** strategy to alleviate curse-of-dimensionality for large-scale black-box
       optimization. Refer to `[Panait et al., 2008, JMLR] <https://jmlr.org/papers/volume9/panait08a/panait08a.pdf>`_
       for convergence analyses and e.g. `[Gomez et al.,, 2008, JMLR] <https://www.jmlr.org/papers/v9/gomez08a.html>`_
       for state-of-the-art *neuroevolution* applications from Schmidhuber and/or Miikkulainen's lab.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'n_individuals' - number of individuals/samples, aka population size (`int`, default: `100`).

    References
    ----------
    Gomez, F., Schmidhuber, J. and Miikkulainen, R., 2008.
    Accelerated neural evolution through cooperatively coevolved synapses.
    Journal of Machine Learning Research, 9(31), pp.937-965.
    https://www.jmlr.org/papers/v9/gomez08a.html

    Panait, L., Tuyls, K. and Luke, S., 2008.
    Theoretical advantages of lenient learners: An evolutionary game theoretic perspective.
    Journal of Machine Learning Research, 9, pp.423-457.
    https://jmlr.org/papers/volume9/panait08a/panait08a.pdf

    Schmidhuber, J., Wierstra, D., Gagliolo, M. and Gomez, F., 2007.
    Training recurrent networks by evolino.
    Neural Computation, 19(3), pp.757-779.
    https://direct.mit.edu/neco/article-abstract/19/3/757/7156/Training-Recurrent-Networks-by-Evolino

    Gomez, F.J. and Schmidhuber, J., 2005, June.
    Co-evolving recurrent neurons learn deep memory POMDPs.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 491-498). ACM.
    https://dl.acm.org/doi/10.1145/1068009.1068092

    Fan, J., Lau, R. and Miikkulainen, R., 2003.
    Utilizing domain knowledge in neuroevolution.
    In International Conference on Machine Learning (pp. 170-177).
    https://www.aaai.org/Library/ICML/2003/icml03-025.php

    Potter, M.A. and De Jong, K.A., 2000.
    Cooperative coevolution: An architecture for evolving coadapted subcomponents.
    Evolutionary Computation, 8(1), pp.1-29.
    https://direct.mit.edu/evco/article/8/1/1/859/Cooperative-Coevolution-An-Architecture-for

    Gomez, F.J. and Miikkulainen, R., 1999, July.
    Solving non-Markovian control tasks with neuroevolution.
    In Proceedings of International Joint Conference on Artificial Intelligence (pp. 1356-1361).
    https://www.ijcai.org/Proceedings/99-2/Papers/097.pdf

    Moriarty, D.E. and Mikkulainen, R., 1996.
    Efficient reinforcement learning through symbiotic evolution.
    Machine Learning, 22(1), pp.11-32.
    https://link.springer.com/article/10.1023/A:1018004120707

    Moriarty, D.E. and Miikkulainen, R., 1995.
    Efficient learning from delayed rewards through symbiotic evolution.
    In International Conference on Machine Learning (pp. 396-404). Morgan Kaufmann.
    https://www.sciencedirect.com/science/article/pii/B9781558603776500566

    Potter, M.A. and De Jong, K.A., 1994, October.
    A cooperative coevolutionary approach to function optimization.
    In International Conference on Parallel Problem Solving from Nature (pp. 249-257).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/3-540-58484-6_269
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self.n_individuals = options.get('n_individuals', 100)  # number of individuals/samples, aka population size
        assert self.n_individuals > 0
        self._n_generations = 0  # initial number of generations (cycles)
        self._printed_evaluations = self.n_function_evaluations  # for printing

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        pass

    def _print_verbose_info(self, fitness, y, is_print=False):
        if y is not None:
            if self.saving_fitness:
                if not np.isscalar(y):
                    fitness.extend(y)
                else:
                    fitness.append(y)
        if self.verbose:
            is_verbose = self._printed_evaluations != self.n_function_evaluations  # to avoid repeated printing
            is_verbose_1 = (not self._n_generations % self.verbose) and is_verbose
            is_verbose_2 = self.termination_signal > 0 and is_verbose
            is_verbose_3 = is_print and is_verbose
            if is_verbose_1 or is_verbose_2 or is_verbose_3:
                info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
                print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))
                self._printed_evaluations = self.n_function_evaluations

    def _collect(self, fitness, y=None):
        self._print_verbose_info(fitness, y)
        results = Optimizer._collect(self, fitness)
        results['_n_generations'] = self._n_generations
        return results



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.pso.spso import PSO


class CPSO(PSO):
    """Cooperative Particle Swarm Optimizer (CPSO).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, aka number of particles (`int`, default: `20`),
                * 'cognition'     - cognitive learning rate (`float`, default: `1.49`),
                * 'society'       - social learning rate (`float`, default: `1.49`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.2`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.pso.cpso import CPSO
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> cpso = CPSO(problem, options)  # initialize the optimizer class
       >>> results = cpso.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CPSO: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CPSO: 5000, 0.3085868239334274

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/27nx42rm>`_ for more details.

    Attributes
    ----------
    cognition     : `float`
                    cognitive learning rate, aka acceleration coefficient.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.
    n_individuals : `int`
                    swarm (population) size, aka number of particles.
    society       : `float`
                    social learning rate, aka acceleration coefficient.

    References
    ----------
    Van den Bergh, F. and Engelbrecht, A.P., 2004.
    A cooperative approach to particle swarm optimization.
    IEEE Transactions on Evolutionary Computation, 8(3), pp.225-239.
    https://ieeexplore.ieee.org/document/1304845
    """
    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)
        self.cognition = options.get('cognition', 1.49)  # cognitive learning rate
        assert self.cognition >= 0.0
        self.society = options.get('society', 1.49)  # social learning rate
        assert self.society >= 0.0
        self._max_generations = np.ceil(self.max_function_evaluations/(self.n_individuals*self.ndim_problem))
        self._w = 1.0 - (np.arange(self._max_generations) + 1.0)/self._max_generations  # from 1.0 to 0.0

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, n_x=None, args=None):
        fitness = []
        for j in range(self.ndim_problem):
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x, fitness
            cognition_rand = self.rng_optimization.uniform(size=(self.n_individuals, self.ndim_problem))
            society_rand = self.rng_optimization.uniform(size=(self.n_individuals, self.ndim_problem))
            for i in range(self.n_individuals):
                if self._check_terminations():
                    return v, x, y, p_x, p_y, n_x, fitness
                n_x[i, j] = p_x[np.argmin(p_y), j]
                v[i, j] = (self._w[min(self._n_generations, len(self._w))]*v[i, j] +
                           self.cognition*cognition_rand[i, j]*(p_x[i, j] - x[i, j]) +
                           self.society*society_rand[i, j]*(n_x[i, j] - x[i, j]))  # velocity update
                v[i, j] = np.clip(v[i, j], self._min_v[j], self._max_v[j])
                x[i, j] += v[i, j]  # position update
                xx = np.copy(self.best_so_far_x)
                xx[j] = x[i, j]
                y[i] = self._evaluate_fitness(xx, args)
                fitness.append(y[i])
                if y[i] < p_y[i]:  # online update
                    p_x[i, j], p_y[i] = x[i, j], y[i]
        self._n_generations += 1
        return v, x, y, p_x, p_y, n_x, fitness

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        v, x, y, p_x, p_y, n_x = self.initialize(args)
        yy = y  # only for printing
        while not self.termination_signal:
            self._print_verbose_info(fitness, yy)
            v, x, y, p_x, p_y, n_x, yy = self.iterate(v, x, y, p_x, p_y, n_x, args)
        return self._collect(fitness, yy)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.ga.ga import GA


class G3PCX(GA):
    """Generalized Generation Gap with Parent-Centric Recombination (G3PCX).

    .. note:: Originally `G3PCX` was proposed to scale up the efficiency of `GA` mainly by Deb,
       `the recipient of IEEE Evolutionary Computation Pioneer Award 2018 <https://tinyurl.com/456as566>`_.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - population size (`int`, default: `100`),
                * 'n_parents'     - parent size (`int`, default: `3`),
                * 'n_offsprings'  - offspring size (`int`, default: `2`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ga.g3pcx import G3PCX
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> g3pcx = G3PCX(problem, options)  # initialize the optimizer class
       >>> results = g3pcx.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"G3PCX: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       G3PCX: 5000, 0.0

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/yc39x6d5>`_ for more details.

    Attributes
    ----------
    n_individuals : `int`
                    population size.
    n_offsprings  : `int`
                    offspring size.
    n_parents     : `int`
                    parent size.

    References
    ----------
    https://www.egr.msu.edu/~kdeb/codes/g3pcx/g3pcx.tar    (See the original C source code.)

    https://pymoo.org/algorithms/soo/g3pcx.html

    Deb, K., Anand, A. and Joshi, D., 2002.
    A computationally efficient evolutionary algorithm for real-parameter optimization.
    Evolutionary Computation, 10(4), pp.371-395.
    https://direct.mit.edu/evco/article-abstract/10/4/371/1136/A-Computationally-Efficient-Evolutionary-Algorithm
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.n_offsprings = options.get('n_offsprings', 2)
        assert self.n_offsprings > 0
        self.n_parents = options.get('n_parents', 3)
        assert self.n_parents > 0
        self._std_pcx_1 = options.get('_std_pcx_1', 0.1)
        self._std_pcx_2 = options.get('_std_pcx_2', 0.1)
        self._elitist = None  # index of elitist

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y

    def iterate(self, x=None, y=None, args=None):
        self._elitist, fitness = np.argmin(y), []
        # (Step 1:) select the best and `self.n_parents - 1` other parents randomly from the population
        parents = self.rng_optimization.choice(self.n_individuals, size=self.n_parents, replace=False)
        if self._elitist not in parents:  # to ensure that elitist is always included
            parents[0] = self._elitist
        # (Step 2:) generate offspring from the chosen parents using a recombination scheme
        xx, yy = np.empty((self.n_offsprings, self.ndim_problem)), np.empty((self.n_offsprings,))
        g = np.mean(x[parents], axis=0)  # mean vector of the chosen parents
        for i in range(self.n_offsprings):
            if self._check_terminations():
                break
            p = self._elitist  # for faster local convergence
            d = g - x[p]
            d_norm = np.linalg.norm(d)
            d_mean = np.empty((self.n_parents - 1,))
            diff = np.empty((self.n_parents - 1, self.ndim_problem))  # for distance computation
            for ii, j in enumerate(parents[1:]):
                diff[ii] = x[j] - x[p]  # distance from one parent
            for ii in range(self.n_parents - 1):
                d_mean[ii] = np.linalg.norm(diff[ii])*np.sqrt(
                    1.0 - np.power(np.dot(diff[ii], d)/(np.linalg.norm(diff[ii])*d_norm), 2))
            d_mean = np.mean(d_mean)  # average of perpendicular distances
            orth = self._std_pcx_2*d_mean*self.rng_optimization.standard_normal((self.ndim_problem,))
            orth = orth - (np.dot(orth, d)*d)/np.power(d_norm, 2)
            xx[i] = x[p] + self._std_pcx_1*self.rng_optimization.standard_normal()*d + orth
            yy[i] = self._evaluate_fitness(xx[i], args)
            fitness.append(yy[i])
        # (Step 3:) choose two parents at random from the population
        offsprings = self.rng_optimization.choice(self.n_individuals, size=2, replace=False)
        # (Step 4:) from a combined subpopulation of two chosen parents and created offspring, choose
        #   the best two solutions and replace the chosen two parents (in Step 3) with these solutions
        xx, yy = np.vstack((xx, x[offsprings])), np.hstack((yy, y[offsprings]))
        x[offsprings], y[offsprings] = xx[np.argsort(yy)[:2]], yy[np.argsort(yy)[:2]]
        self._n_generations += 1
        return fitness

    def _collect(self, fitness, y=None):
        self._print_verbose_info(fitness, y)
        for i in range(1, len(fitness)):  # to avoid `np.nan`
            if np.isnan(fitness[i]):
                fitness[i] = fitness[i - 1]
        results = Optimizer._collect(self, fitness)
        results['_n_generations'] = self._n_generations
        return results

    def optimize(self, fitness_function=None, args=None):
        fitness = GA.optimize(self, fitness_function)
        x, y = self.initialize(args)
        yy = y  # only for printing
        while not self._check_terminations():
            self._print_verbose_info(fitness, yy)
            yy = self.iterate(x, y, args)
        return self._collect(fitness, yy)



"""Repeat the following paper for `COSYNE`:
    Gomez, F., Schmidhuber, J. and Miikkulainen, R., 2008.
    Accelerated neural evolution through cooperatively coevolved synapses.
    Journal of Machine Learning Research, 9(31), pp.937-965.
    https://jmlr.org/papers/v9/gomez08a.html

    We notice that the EvoTorch library (recently developed in the company of two original authors)
      provides an open-source implementation for it:
    https://docs.evotorch.ai/v0.3.0/reference/evotorch/algorithms/ga/#evotorch.algorithms.ga.Cosyne

    In our implementation, we refer to this open-source implementation rather than the original paper,
    but with a slight simplification for the *permutation* operator.

    Luckily our Python code could repeat the data reported in the reference library *well*.
    Therefore, we argue that its repeatability could be **well-documented**.



    The reference Python code based on EvoTorch is given below:
    -----------------------------------------------------------
    import torch
    from evotorch import Problem
    from evotorch.algorithms import ga
    from evotorch.logging import StdOutLogger


    def norm(x: torch.Tensor) -> torch.Tensor:
        return (torch.linalg.norm(x, dim=-1)).pow(2)


    problem = Problem('min', norm, initial_bounds=(-5.0, 5.0), solution_length=1000)
    searcher = ga.Cosyne(problem, popsize=100, tournament_size=10, mutation_probability=1.0, mutation_stdev=1.0)
    logger = StdOutLogger(searcher)
    searcher.run(num_generations=3000)
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import sphere
from pypop7.optimizers.cc.cosyne import COSYNE as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 1000
    for f in [sphere]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -5*np.ones((ndim_problem,)),
                   'upper_boundary': 5*np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 150*3000 + 100,
                   'sigma': 1.0,
                   'seed_rng': 0}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)  # 7368.388082444299 vs 7388.05517578125 (from EvoTorch library)
        print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



import numpy as np  # engine for numerical computing
from scipy.stats import levy_stable

from pypop7.optimizers.ep.cep import CEP


class LEP(CEP):
    """Lévy distribution based Evolutionary Programming (LEP).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'          - initial global step-size, aka mutation strength (`float`),
                * 'n_individuals'  - number of offspring, aka offspring population size (`int`, default: `100`),
                * 'q'              - number of opponents for pairwise comparisons (`int`, default: `10`),
                * 'tau'            - learning rate of individual step-sizes self-adaptation (`float`, default:
                  `1.0/np.sqrt(2.0*np.sqrt(problem['ndim_problem']))`),
                * 'tau_apostrophe' - learning rate of individual step-sizes self-adaptation (`float`, default:
                  `1.0/np.sqrt(2.0*problem['ndim_problem'])`.

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy  # engine for numerical computing
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ep.lep import LEP
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'sigma': 0.1}
       >>> lep = LEP(problem, options)  # initialize the optimizer class
       >>> results = lep.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"LEP: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       LEP: 5000, 0.0359694938656471

    For its correctness checking, refer to `this code-based repeatability report
    <https://tinyurl.com/2dc6ym6j>`_ for more details.

    Attributes
    ----------
    n_individuals  : `int`
                     number of offspring, aka offspring population size.
    q              : `int`
                     number of opponents for pairwise comparisons.
    sigma          : `float`
                     initial global step-size, aka mutation strength.
    tau            : `float`
                     learning rate of individual step-sizes self-adaptation.
    tau_apostrophe : `float`
                     learning rate of individual step-sizes self-adaptation.

    References
    ----------
    Lee, C.Y. and Yao, X., 2004.
    Evolutionary programming using mutations based on the Lévy probability distribution.
    IEEE Transactions on Evolutionary Computation, 8(1), pp.1-13.
    https://ieeexplore.ieee.org/document/1266370
    """
    def __init__(self, problem, options):
        CEP.__init__(self, problem, options)

    def iterate(self, x=None, sigmas=None, y=None, xx=None, ss=None, yy=None, args=None):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, sigmas, y, xx, ss, yy
            ss[i] = sigmas[i]*np.exp(self.tau_apostrophe*self.rng_optimization.standard_normal(
                size=(self.ndim_problem,)) + self.tau*self.rng_optimization.standard_normal(
                size=(self.ndim_problem,)))
            xx[i] = x[i] + ss[i]*levy_stable.rvs(alpha=1.8, beta=1, size=(self.ndim_problem,),
                                                 random_state=self.rng_optimization)
            yy[i] = self._evaluate_fitness(xx[i], args)
        new_x = np.vstack((xx, x))
        new_sigmas = np.vstack((ss, sigmas))
        new_y = np.hstack((yy, y))
        n_win = np.zeros((2*self.n_individuals,))  # number of win
        for i in range(2*self.n_individuals):
            for j in self.rng_optimization.choice([k for k in range(2*self.n_individuals) if k != i],
                                                  size=self.q, replace=False):
                if new_y[i] < new_y[j]:
                    n_win[i] += 1
        order = np.argsort(-n_win)[:self.n_individuals]
        x[:self.n_individuals] = new_x[order]
        sigmas[:self.n_individuals] = new_sigmas[order]
        y[:self.n_individuals] = new_y[order]
        self._n_generations += 1
        return x, sigmas, y, xx, ss, yy



import numpy as np

from pypop7.optimizers.es.es import ES


class LMCMA(ES):
    """Limited-Memory Covariance Matrix Adaptation (LMCMA).

    .. note:: Currently `LMCMA` is a **State-Of-The-Art (SOTA)** variant of `CMA-ES` designed especially for large-scale
       black-box optimization. Inspired by `L-BFGS` (a well-established *second-order* gradient-based optimizer),
       it stores only *m* direction vectors to reconstruct the covariance matirx on-the-fly, resulting in **O(mn)**
       time complexity w.r.t. each sampling, where *m=O(log(n))* and *n* is the dimensionality of objective function.

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'm'             - number of direction vectors (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'base_m'        - base number of direction vectors (`int`, default: `4`),
                * 'period'        - update period (`int`, default:
                  `int(np.maximum(1, np.log(problem['ndim_problem'])))`),
                * 'n_steps'       - target number of generations between vectors (`int`, default:
                  `problem['ndim_problem']`),
                * 'c_c'           - learning rate for evolution path update (`float`, default:
                  `0.5/np.sqrt(problem['ndim_problem'])`),
                * 'c_1'           - learning rate for covariance matrix adaptation (`float`, default:
                  `1.0/(10.0*np.log(problem['ndim_problem'] + 1.0))`),
                * 'c_s'           - learning rate for population success rule (`float`, default: `0.3`),
                * 'd_s'           - changing rate for population success rule (`float`, default: `1.0`),
                * 'z_star'        - target success rate for population success rule (`float`, default: `0.3`),
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.lmcma import LMCMA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 1000,
       ...            'lower_boundary': -5*numpy.ones((1000,)),
       ...            'upper_boundary': 5*numpy.ones((1000,))}
       >>> options = {'max_function_evaluations': 1e5*1000,  # set optimizer options
       ...            'fitness_threshold': 1e-8,
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((1000,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> lmcma = LMCMA(problem, options)  # initialize the optimizer class
       >>> results = lmcma.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"LMCMA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       LMCMA: 2482983, 9.998653039116044e-09

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/24jnfhbs>`_ for more details.

    Attributes
    ----------
    base_m        : `int`
                    base number of direction vectors.
    c_c           : `float`
                    learning rate for evolution path update.
    c_s           : `float`
                    learning rate for population success rule.
    c_1           : `float`
                    learning rate for covariance matrix adaptation.
    d_s           : `float`
                    changing rate for population success rule.
    m             : `int`
                    number of direction vectors.
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    n_steps       : `int`
                    target number of generations between vectors.
    period        : `int`
                    update period.
    sigma         : `float`
                    final global step-size, aka mutation strength.
    z_star        : `float`
                    target success rate for population success rule.

    References
    ----------
    Loshchilov, I., 2017.
    LM-CMA: An alternative to L-BFGS for large-scale black box optimization.
    Evolutionary Computation, 25(1), pp.143-171.
    https://direct.mit.edu/evco/article-abstract/25/1/143/1041/LM-CMA-An-Alternative-to-L-BFGS-for-Large-Scale
    (See Algorithm 7 for details.)

    See the official C++ version from Loshchilov, which provides an interface for Matlab users:
    https://sites.google.com/site/ecjlmcma/
    (Unfortunately, this website link appears to be not available now.)
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.m = options.get('m', 4 + int(3*np.log(self.ndim_problem)))  # number of direction vectors
        self.base_m = options.get('base_m', 4)  # base number of direction vectors
        self.period = options.get('period', int(np.maximum(1, np.log(self.ndim_problem))))  # update period
        self.n_steps = options.get('n_steps', self.ndim_problem)  # target number of generations between vectors
        self.c_c = options.get('c_c', 0.5/np.sqrt(self.ndim_problem))  # learning rate for evolution path
        self.c_1 = options.get('c_1', 1.0/(10.0*np.log(self.ndim_problem + 1.0)))
        self.c_s = options.get('c_s', 0.3)  # learning rate for population success rule (PSR)
        self.d_s = options.get('d_s', 1.0)  # changing rate for PSR
        self.z_star = options.get('z_star', 0.3)  # target success rate for PSR
        self._a = np.sqrt(1.0 - self.c_1)
        self._c = 1.0/np.sqrt(1.0 - self.c_1)
        self._bd_1 = np.sqrt(1.0 - self.c_1)
        self._bd_2 = self.c_1/(1.0 - self.c_1)
        self._p_c_1 = 1.0 - self.c_c
        self._p_c_2 = None
        self._j = None
        self._l = None
        self._it = None
        self._rr = None  # for PSR

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        s = 0.0  # for PSR
        vm = np.empty((self.m, self.ndim_problem))
        pm = np.empty((self.m, self.ndim_problem))
        b = np.empty((self.m,))
        d = np.empty((self.m,))
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._p_c_2 = np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)
        self._j = [None]*self.m
        self._l = [None]*self.m
        self._it = 0
        self._rr = np.arange(self.n_individuals*2, 0, -1) - 1
        return mean, x, p_c, s, vm, pm, b, d, y

    def _rademacher(self):
        """Sampling from Rademacher distribution."""
        random = self.rng_optimization.integers(2, size=(self.ndim_problem,))
        random[random == 0] = -1
        return np.double(random)

    def _a_z(self, z=None, pm=None, vm=None, b=None, start=None, it=None):
        """Az(): Cholesky factor-vector update."""
        x = np.copy(z)
        for t in range(start, it):
            x = self._a*x + b[self._j[t]]*np.dot(vm[self._j[t]], z)*pm[self._j[t]]
        return x

    def iterate(self, mean=None, x=None, pm=None, vm=None, y=None, b=None, args=None):
        sign, a_z = 1, np.empty((self.ndim_problem,))  # for mirrored sampling
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            if sign == 1:  # SelectSubst(): direction vectors selection
                base_m = (10.0*self.base_m if k == 0 else self.base_m)*np.abs(
                    self.rng_optimization.standard_normal())
                base_m = float(self._it if base_m > self._it else base_m)
                a_z = self._a_z(self._rademacher(), pm, vm, b,
                                int(self._it - base_m) if self._it > 1 else 0, self._it)
            x[k] = mean + sign*self.sigma*a_z
            y[k] = self._evaluate_fitness(x[k], args)
            sign *= -1  # sampling in the opposite direction for mirrored sampling
        return x, y

    def _a_inv_z(self, v=None, vm=None, d=None, i=None):
        """Ainvz(): inverse Cholesky factor-vector update."""
        x = np.copy(v)
        for t in range(0, i):
            x = self._c*x - d[self._j[t]]*np.dot(vm[self._j[t]], x)*vm[self._j[t]]
        return x

    def update_distribution(self, mean=None, x=None, p_c=None, s=None, vm=None,
                            pm=None, b=None, d=None, y=None, y_bak=None):
        mean_bak = np.dot(self._w, x[np.argsort(y)[:self.n_parents]])
        p_c = self._p_c_1*p_c + self._p_c_2*(mean_bak - mean)/self.sigma
        # select and store direction vectors - to preserve a certain temporal distance in terms of
        #   number of generations between the stored direction vectors
        if self._n_generations % self.period == 0:  # temporal distance
            _n_generations = int(self._n_generations/self.period)  # temporal distance
            i_min = 1  # index of the first vector that will be replaced by the new one
            # the higher is the index of `self._j`, the more recent is the corresponding direction vector
            if _n_generations < self.m:
                self._j[_n_generations] = _n_generations
            else:
                if self.m > 1:
                    # find a pair of consecutively saved vectors with the distance between them
                    #   closest to a target distance
                    d_min = (self._l[self._j[1]] - self._l[self._j[0]]) - self.n_steps
                    for j in range(2, self.m):
                        d_cur = (self._l[self._j[j]] - self._l[self._j[j - 1]]) - self.n_steps
                        if d_cur < d_min:
                            d_min, i_min = d_cur, j
                    i_min = 0 if d_min >= 0 else i_min
                    updated = self._j[i_min]
                    for j in range(i_min, self.m - 1):
                        self._j[j] = self._j[j + 1]
                    self._j[self.m - 1] = updated
            self._it = int(np.minimum(self.m, _n_generations + 1))
            self._l[self._j[self._it - 1]] = _n_generations*self.period
            pm[self._j[self._it - 1]] = p_c
            for i in range(0 if i_min == 1 else i_min, self._it):
                vm[self._j[i]] = self._a_inv_z(pm[self._j[i]], vm, d, i)
                v_n = np.dot(vm[self._j[i]], vm[self._j[i]])
                bd_3 = np.sqrt(1.0 + self._bd_2*v_n)
                b[self._j[i]] = self._a/v_n*(bd_3 - 1.0)
                d[self._j[i]] = self._c/v_n*(1.0 - 1.0/bd_3)
        if self._n_generations > 0:  # for PSR
            r = np.argsort(np.hstack((y, y_bak)))
            z_psr = np.sum(self._rr[r < self.n_individuals] - self._rr[r >= self.n_individuals])
            z_psr = z_psr/np.power(self.n_individuals, 2) - self.z_star
            s = (1.0 - self.c_s)*s + self.c_s*z_psr
            self.sigma *= np.exp(s/self.d_s)
        return mean_bak, p_c, s, vm, pm, b, d

    def restart_reinitialize(self, mean=None, x=None, p_c=None, s=None, vm=None,
                             pm=None, b=None, d=None, y=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            mean, x, p_c, s, vm, pm, b, d, y = self.initialize(True)
        return mean, x, p_c, s, vm, pm, b, d, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, x, p_c, s, vm, pm, b, d, y = self.initialize()
        while not self.termination_signal:
            y_bak = np.copy(y)
            # sample and evaluate offspring population
            x, y = self.iterate(mean, x, pm, vm, y, b, args)
            if self._check_terminations():
                break
            mean, p_c, s, vm, pm, b, d = self.update_distribution(
                mean, x, p_c, s, vm, pm, b, d, y, y_bak)
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            mean, x, p_c, s, vm, pm, b, d, y = self.restart_reinitialize(
                mean, x, p_c, s, vm, pm, b, d, y)
        results = self._collect(fitness, y, mean)
        results['p_c'] = p_c
        results['s'] = s
        return results



import numpy as np

from pypop7.optimizers.es.es import ES


class VDCMA(ES):
    """Linear Covariance Matrix Adaptation (VDCMA).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.vdcma import VDCMA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> vdcma = VDCMA(problem, options)  # initialize the optimizer class
       >>> results = vdcma.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"VDCMA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       VDCMA: 5000, 7.116226375179302e-18

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/3e838zd5>`_ for more details.

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    sigma         : `float`
                    final global step-size, aka mutation strength.

    References
    ----------
    Akimoto, Y., Auger, A. and Hansen, N., 2014, July.
    Comparison-based natural gradient optimization in high dimension.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 373-380). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598258

    See the official Python version from Prof. Akimoto:
    https://gist.github.com/youheiakimoto/08b95b52dfbf8832afc71dfff3aed6c8
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        self.c_factor = options.get('c_factor', np.maximum((self.ndim_problem - 5.0)/6.0, 0.5))
        self.c_c, self.c_1, self.c_mu, self.c_s = None, None, None, None
        self.d_s = None
        self._v_n, self._v_2, self._v_, self._v_p2 = None, None, None, None

    def initialize(self, is_restart=False):
        self.c_c = self.options.get('c_c', (4.0 + self._mu_eff/self.ndim_problem)/(
                self.ndim_problem + 4.0 + 2.0*self._mu_eff/self.ndim_problem))
        self.c_1 = self.options.get('c_1', self.c_factor*2.0/(
                np.square(self.ndim_problem + 1.3) + self._mu_eff))
        self.c_mu = self.options.get('c_mu', np.minimum(1.0 - self.c_1, self.c_factor*2.0*(
                self._mu_eff - 2.0 + 1.0/self._mu_eff)/(np.square(self.ndim_problem + 2.0) + self._mu_eff)))
        self.c_s = self.options.get('c_s', 1.0/(2.0*np.sqrt(self.ndim_problem/self._mu_eff) + 1.0))
        self.d_s = self.options.get('d_s', 1.0 + self.c_s + 2.0*np.maximum(0.0, np.sqrt(
            (self._mu_eff - 1.0)/(self.ndim_problem + 1.0)) - 1.0))
        d = np.ones((self.ndim_problem,))  # diagonal vector of sampling distribution
        # set principal search direction (vector) of sampling distribution
        v = self.rng_optimization.standard_normal((self.ndim_problem,))/np.sqrt(self.ndim_problem)
        p_s = np.zeros((self.ndim_problem,))  # evolution path for step-size adaptation (MCSA)
        p_c = np.zeros((self.ndim_problem,))  # evolution path for covariance matrix adaptation (CMA)
        self._v_n = np.linalg.norm(v)
        self._v_2 = np.square(self._v_n)
        self._v_ = v/self._v_n
        self._v_p2 = np.square(self._v_)
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        zz = np.empty((self.n_individuals, self.ndim_problem))  # search directions
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return d, v, p_s, p_c, z, zz, x, mean, y

    def iterate(self, d=None, z=None, zz=None, x=None, mean=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return z, zz, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            zz[k] = d*(z[k] + (np.sqrt(1.0 + self._v_2) - 1.0)*(np.dot(z[k], self._v_)*self._v_))
            x[k] = mean + self.sigma*zz[k]
            y[k] = self._evaluate_fitness(x[k], args)
        return z, zz, x, y

    def _p_q(self, zz, w=0):
        zz_v_ = np.dot(zz, self._v_)
        if isinstance(w, int) and w == 0:
            p = np.square(zz) - self._v_2/(1.0 + self._v_2)*(zz_v_*(zz*self._v_)) - 1.0
            q = zz_v_*zz - ((np.square(zz_v_) + 1.0 + self._v_2)/2.0)*self._v_
        else:
            p = np.dot(w, np.square(zz) - self._v_2/(1.0 + self._v_2)*(zz_v_*(zz*self._v_).T).T - 1.0)
            q = np.dot(w, (zz_v_*zz.T).T - np.outer((np.square(zz_v_) + 1.0 + self._v_2)/2.0, self._v_))
        return p, q

    def _update_distribution(self, d=None, v=None, p_s=None, p_c=None, zz=None, x=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        # update mean
        mean = np.dot(self._w, x[order])
        # update global step-size
        z = np.dot(self._w, zz[order])/d
        z += (1.0/np.sqrt(1.0 + self._v_2) - 1.0)*np.dot(z, self._v_)*self._v_
        p_s = (1.0 - self.c_s)*p_s + np.sqrt(self.c_s*(2.0 - self.c_s)*self._mu_eff)*z
        p_s_2 = np.dot(p_s, p_s)
        self.sigma *= np.exp(self.c_s/self.d_s*(np.sqrt(p_s_2)/self._e_chi - 1.0))
        # update restricted covariance matrix (d, v)
        h_s = p_s_2 < (2.0 + 4.0/(self.ndim_problem + 1.0))*self.ndim_problem
        p_c = (1.0 - self.c_c)*p_c + h_s*np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)*np.dot(self._w, zz[order])
        gamma = 1.0/np.sqrt(1.0 + self._v_2)
        alpha = np.sqrt(np.square(self._v_2) + (1.0 + self._v_2)/np.max(self._v_p2)*(2.0 - gamma))/(2.0 + self._v_2)
        if alpha < 1.0:
            beta = (4.0 - (2.0 - gamma)/np.max(self._v_p2))/np.square(1.0 + 2.0/self._v_2)
        else:
            alpha, beta = 1.0, 0.0
        b = 2.0*np.square(alpha) - beta
        a = 2.0 - (b + 2.0*np.square(alpha))*self._v_p2
        _v_p2_a = self._v_p2/a
        if self.c_mu == 0:
            p_mu, q_mu = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,))
        else:
            p_mu, q_mu = self._p_q(zz[order]/d, self._w)
        if self.c_1 == 0:
            p_1, q_1 = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,))
        else:
            p_1, q_1 = self._p_q(p_c/d)
        p = self.c_mu*p_mu + h_s*self.c_1*p_1
        q = self.c_mu*q_mu + h_s*self.c_1*q_1
        if self.c_mu + self.c_1 > 0:
            r = p - alpha/(1.0 + self._v_2)*((2.0 + self._v_2)*(
                    q*self._v_) - self._v_2*np.dot(self._v_, q)*self._v_p2)
            s = r/a - b*np.dot(r, _v_p2_a)/(1.0 + b*np.dot(self._v_p2, _v_p2_a))*_v_p2_a
            ng_v = q/self._v_n - alpha/self._v_n*((2.0 + self._v_2)*(
                    self._v_*s) - np.dot(s, np.square(self._v_))*self._v_)
            ng_d = d*s
            up_factor = np.minimum(np.minimum(1.0, 0.7*self._v_n/np.sqrt(np.dot(ng_v, ng_v))),
                                   0.7*(d/np.min(np.abs(ng_d))))
        else:
            ng_v, ng_d, up_factor = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,)), 1.0
        v += up_factor*ng_v
        d += up_factor*ng_d
        self._v_n = np.linalg.norm(v)
        self._v_2 = np.square(self._v_n)
        self._v_ = v/self._v_n
        self._v_p2 = np.square(self._v_)
        return mean, p_s, p_c, v, d

    def restart_reinitialize(self, d=None, v=None, p_s=None, p_c=None, z=None, zz=None, x=None, mean=None, y=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            d, v, p_s, p_c, z, zz, x, mean, y = self.initialize(True)
        return d, v, p_s, p_c, z, zz, x, mean, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        d, v, p_s, p_c, z, zz, x, mean, y = self.initialize()
        while not self.termination_signal:
            # sample and evaluate offspring population
            z, zz, x, y = self.iterate(d, z, zz, x, mean, y, args)
            if self._check_terminations():
                break
            self._print_verbose_info(fitness, y)
            mean, p_s, p_c, v, d = self._update_distribution(d, v, p_s, p_c, zz, x, y)
            self._n_generations += 1
            d, v, p_s, p_c, z, zz, x, mean, y = self.restart_reinitialize(
                d, v, p_s, p_c, z, zz, x, mean, y)
        results = self._collect(fitness, y, mean)
        results['d'] = d
        results['v'] = v
        return results



from pypop7.optimizers.es.maes import MAES


class FMAES(MAES):
    """Fast Matrix Adaptation Evolution Strategy (FMAES).

    .. note:: `FMAES` is a *more efficient* implementation of `MAES` with *quadractic* time complexity w.r.t. each
       sampling, which replaces the computationally expensive matrix-matrix multiplication (*cubic time complexity*)
       with the combination of matrix-matrix addition and matrix-vector multiplication (*quadractic time complexity*)
       for transformation matrix adaptation. It is **highly recommended** to first attempt more advanced ES variants
       (e.g., `LMCMA`, `LMMAES`) for large-scale black-box optimization, since `FMAES` still has a computationally
       intensive *quadratic* time complexity (w.r.t. each sampling).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer `FMAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.fmaes import FMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> fmaes = FMAES(problem, options)  # initialize the optimizer class
       >>> results = fmaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"FMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       FMAES: 5000, 2.1296244414852865e-19

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/37ews6h4>`_ for more details.

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    sigma         : `float`
                    final global step-size, aka mutation strength.

    References
    ----------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700).
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115

    See the official Matlab version from Prof. Beyer:
    https://homepages.fhv.at/hgb/downloads/ForDistributionFastMAES.tar
    """
    def __init__(self, problem, options):
        options['_fast_version'] = True  # mandatory setting
        MAES.__init__(self, problem, options)



"""This is a simple demo for the popular Lennard-Jones clustering optimization problem
  from the chemistry field:

  https://esa.github.io/pagmo2/docs/cpp/problems/lennard_jones.html
  https://esa.github.io/pygmo2/install.html
"""
import pygmo as pg  # need to be installed: https://esa.github.io/pygmo2/install.html
from pypop7.optimizers.de.cde import CDE  # https://pypop.readthedocs.io/en/latest/de/cde.html
from pypop7.optimizers.de.jade import JADE  # https://pypop.readthedocs.io/en/latest/de/jade.html
import seaborn as sns
import matplotlib.pyplot as plt


# see https://esa.github.io/pagmo2/docs/cpp/problems/lennard_jones.html for the fitness function
prob = pg.problem(pg.lennard_jones(150))
print(prob)  # 444-dimensional


def energy_func(x):  # wrapper to obtain fitness of type `float`
    return float(prob.fitness(x))


if __name__ == '__main__':
    results = []  # to save all optimization results from different optimizers
    for DE in [CDE, JADE]:
        problem = {'fitness_function': energy_func,
                   'ndim_problem': 444,
                   'upper_boundary': prob.get_bounds()[1],
                   'lower_boundary': prob.get_bounds()[0]}
        if DE == JADE:  # for JADE (but not for CDE)
            is_bound = True
        else:
            is_bound = False
        options = {'max_function_evaluations': 400000,
                   'seed_rng': 2022,  # for repeatability
                   'saving_fitness': 1,  # to save all fitness generated during optimization
                   'is_bound': is_bound}
        solver = DE(problem, options)
        results.append(solver.optimize())
        print(results[-1])

    sns.set_theme(style='darkgrid')
    plt.figure()
    for label, res in zip(['CDE', 'JADE'], results):
        # starting from 250000 can avoid excessively high values generated during the early stage
        #   to disrupt convergence curves
        plt.plot(res['fitness'][250000:, 0], res['fitness'][250000:, 1], label=label)

    plt.legend()
    plt.show()



import numpy as np
from scipy.linalg import solve_triangular

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.opoa2015 import cholesky_update


class CCMAES2016(ES):
    """Cholesky-CMA-ES 2016 (CCMAES2016).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.ccmaes2016 import CCMAES2016
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> ccmaes2016 = CCMAES2016(problem, options)  # initialize the optimizer class
       >>> results = ccmaes2016.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CCMAES2016: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CCMAES2016: 5000, 2.614231350522262e-21

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/3kbet9tn>`_ for more details.

    References
    ----------
    Krause, O., Arbonès, D.R. and Igel, C., 2016.
    CMA-ES with optimal covariance update and storage complexity.
    Advances in Neural Information Processing Systems, 29, pp.370-378.
    https://proceedings.neurips.cc/paper/2016/hash/289dff07669d7a23de0ef88d2f7129e7-Abstract.html
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        self.c_s, self.d = None, None
        self.c_c, self.c_1, self.c_mu = None, None, None

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        a = np.diag(np.ones((self.ndim_problem,)))  # cholesky factor
        p_s = np.zeros((self.ndim_problem,))  # evolution path for CSA
        p_c = np.zeros((self.ndim_problem,))  # evolution path for CMA
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self.c_s = self.options.get('c_s', self._mu_eff/(self.ndim_problem + self._mu_eff))
        self.d = self.options.get('d', 1.0 + np.sqrt(self._mu_eff/self.ndim_problem))
        self.c_c = self.options.get('c_c', (4.0 + self._mu_eff/self.ndim_problem)/(
                self.ndim_problem + 4.0 + 2.0*self._mu_eff/self.ndim_problem))
        self.c_1 = self.options.get('c_1', 2.0/(np.square(self.ndim_problem) + self._mu_eff))
        self.c_mu = self.options.get('c_mu', self._mu_eff/(np.square(self.ndim_problem) + self._mu_eff))
        return x, mean, a, p_s, p_c, y

    def iterate(self, x=None, mean=None, a=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            x[k] = mean + self.sigma*np.dot(a, self.rng_optimization.standard_normal((self.ndim_problem,)))
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, a=None, p_s=None, p_c=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        mean_bak = np.dot(self._w, x[order])
        mean_diff = (mean_bak - mean)/self.sigma
        p_c = (1.0 - self.c_c)*p_c + np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)*mean_diff
        p_s = (1.0 - self.c_s)*p_s + np.sqrt(self.c_s*(2.0 - self.c_s)*self._mu_eff)*solve_triangular(
            a, mean_diff, lower=True)
        a *= np.sqrt(1.0 - self.c_1 - self.c_mu)
        a = cholesky_update(a, np.sqrt(self.c_1)*p_c, False)
        for i in range(self.n_parents):
            a = cholesky_update(a, np.sqrt(self.c_mu*self._w[i])*(x[order[i]] - mean)/self.sigma, False)
        self.sigma *= np.exp(self.c_s/self.d*(np.sqrt(np.dot(p_s, p_s))/self._e_chi - 1.0))
        return mean_bak, a, p_s, p_c

    def restart_reinitialize(self, x=None, mean=None, a=None, p_s=None, p_c=None, y=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            x, mean, a, p_s, p_c, y = self.initialize(True)
        return x, mean, a, p_s, p_c, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, a, p_s, p_c, y = self.initialize()
        while not self.termination_signal:
            # sample and evaluate offspring population
            x, y = self.iterate(x, mean, a, y, args)
            if self._check_terminations():
                break
            mean, a, p_s, p_c = self._update_distribution(x, mean, a, p_s, p_c, y)
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            x, mean, a, p_s, p_c, y = self.restart_reinitialize(x, mean, a, p_s, p_c, y)
        results = self._collect(fitness, y, mean)
        results['p_s'] = p_s
        results['p_c'] = p_c
        return results



from pypop7.optimizers.nes.nes import NES  # abstract class for all NES classes
from pypop7.optimizers.nes.sges import SGES
from pypop7.optimizers.nes.ones import ONES
from pypop7.optimizers.nes.enes import ENES
from pypop7.optimizers.nes.xnes import XNES
from pypop7.optimizers.nes.snes import SNES
from pypop7.optimizers.nes.r1nes import R1NES


___all__ = [NES,  # Natural Evolution Strategies (NES)
            SGES,  # Search Gradient-based Evolution Strategy (SGES)
            ONES,  # Original Natural Evolution Strategy (ONES)
            ENES,  # Exact Natural Evolution Strategy (ENES)
            XNES,  # Exponential Natural Evolution Strategies (XNES)
            SNES,  # Separable Natural Evolution Strategies (SNES)
            R1NES]  # Rank-One Natural Evolution Strategies (R1NES)



"""Repeat the following paper for `AEMNA`:
    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5
    (See Chapter 8 Experimental Results in Function Optimization with EDAs in Continuous Domain.)

    Surprisingly, our Python code reported better results than the original paper. However,
    given that it shows very similar performance with its same kind `EMNA`, we still argue that its
    repeatability can be *well-documented*.
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, griewank
from pypop7.optimizers.eda.aemna import AEMNA


if __name__ == '__main__':
    ndim_problem = 10

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 2000,
               'seed_rng': 0,  # undefined in the original paper
               'verbose': 1000}
    aemna = AEMNA(problem, options)
    results = aemna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 8.67464285339613 vs 3263.0010 (from the original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 750,
               'seed_rng': 0,  # undefined in the original paper
               'verbose': 1000}
    aemna = AEMNA(problem, options)
    results = aemna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.59856558868619 vs 12.9407 (from the original paper)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class RS(Optimizer):
    """Random (stochastic) Search (optimization) (RS).

    This is the **abstract** class for all `RS` classes. Please use any of its instantiated subclasses to
    optimize the black-box problem at hand. Recently, all of its state-of-the-art versions adopt the
    **population-based** random sampling strategy for better exploration in the complex search space.

    .. note:: `"The topic of local search was reinvigorated in the early 1990s by surprisingly good results for large
       (combinatorial) problems ... and by the incorporation of randomness, multiple simultaneous searches, and other
       improvements."---[Russell&Norvig, 2022] <http://aima.cs.berkeley.edu/global-index.html>`_

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'x' - initial (starting) point (`array_like`).

    Attributes
    ----------
    x     : `array_like`
            initial (starting) point.

    Methods
    -------

    References
    ----------
    Gao, K. and Sener, O., 2022, June.
    Generalizing Gaussian smoothing for random search.
    In International Conference on Machine Learning (pp. 7077-7101). PMLR.
    https://proceedings.mlr.press/v162/gao22f.html

    Nesterov, Y. and Spokoiny, V., 2017.
    Random gradient-free minimization of convex functions.
    Foundations of Computational Mathematics, 17(2), pp.527-566.
    https://link.springer.com/article/10.1007/s10208-015-9296-2

    Bergstra, J. and Bengio, Y., 2012.
    Random search for hyper-parameter optimization.
    Journal of Machine Learning Research, 13(2).
    https://www.jmlr.org/papers/v13/bergstra12a.html

    Appel, M.J., Labarre, R. and Radulovic, D., 2004.
    On accelerated random search.
    SIAM Journal on Optimization, 14(3), pp.708-731.
    https://epubs.siam.org/doi/abs/10.1137/S105262340240063X

    Schmidhuber, J., Hochreiter, S. and Bengio, Y., 2001.
    Evaluating benchmark problems by random guessing.
    A Field Guide to Dynamical Recurrent Networks, pp.231-235.
    https://ml.jku.at/publications/older/ch9.pdf

    Schmidhuber, J. and Hochreiter, S., 1996.
    Guessing can outperform many long time lag algorithms.
    Technical Report.
    https://www.bioinf.jku.at/publications/older/3204.pdf

    Rastrigin, L.A., 1986.
    Random search as a method for optimization and adaptation.
    In Stochastic Optimization.
    https://link.springer.com/chapter/10.1007/BFb0007129

    Solis, F.J. and Wets, R.J.B., 1981.
    Minimization by random search techniques.
    Mathematics of Operations Research, 6(1), pp.19-30.
    https://pubsonline.informs.org/doi/abs/10.1287/moor.6.1.19

    Schrack, G. and Choit, M., 1976.
    Optimized relative step size random searches.
    Mathematical Programming, 10(1), pp.230-244.
    https://link.springer.com/article/10.1007/BF01580669

    Schumer, M.A. and Steiglitz, K., 1968.
    Adaptive step size random search.
    IEEE Transactions on Automatic Control, 13(3), pp.270-276.
    https://ieeexplore.ieee.org/abstract/document/1098903

    Matyas, J., 1965.
    Random optimization.
    Automation and Remote control, 26(2), pp.246-253.
    https://tinyurl.com/25339c4x
    (*Since it was written originally in Russian, we cannot read it. However, owing to its historical position,
    we still choose to include it here, which causes a nonstandard citation.*)

    Rastrigin, L.A., 1963.
    The convergence of the random search method in the extremal control of a many parameter system.
    Automaton & Remote Control, 24, pp.1337-1342.
    https://tinyurl.com/djfdnpx4

    Brooks, S.H., 1958.
    A discussion of random methods for seeking maxima.
    Operations Research, 6(2), pp.244-251.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.6.2.244
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self.x = options.get('x')  # initial (starting) point
        # reset its default value from 10 to 1000, since typically more iterations can be run
        # for individual-based iterative search
        self.verbose = options.get('verbose', 1000)
        self._n_generations = 0  # number of generations

    def initialize(self):
        raise NotImplementedError

    def iterate(self):  # for each iteration (generation)
        raise NotImplementedError

    def _print_verbose_info(self, fitness, y):
        if self.saving_fitness:
            if not np.isscalar(y):
                fitness.extend(y)
            else:
                fitness.append(y)
        if self.verbose and ((not self._n_generations % self.verbose) or (self.termination_signal > 0)):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect(self, fitness, y=None):
        if y is not None:
            self._print_verbose_info(fitness, y)
        results = Optimizer._collect(self, fitness)
        results['_n_generations'] = self._n_generations
        return results

    def optimize(self, fitness_function=None, args=None):  # for all iterations (generations)
        fitness = Optimizer.optimize(self, fitness_function)
        x = self.initialize()  # starting point
        y = self._evaluate_fitness(x, args)  # fitness of starting point
        while not self._check_terminations():
            self._print_verbose_info(fitness, y)
            x = self.iterate()  # to sample a new point
            y = self._evaluate_fitness(x, args)  # to evaluate the new point
            self._n_generations += 1
        return self._collect(fitness, y)



import numpy as np  # engine for numerical computing
import scipy

from pypop7.optimizers.nes.nes import NES


class XNES(NES):
    """Exponential Natural Evolution Strategies (XNES).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring/descendants, aka offspring population size (`int`),
                * 'n_parents'     - number of parents/ancestors, aka parental population size (`int`),
                * 'mean'          - initial (starting) point (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'sigma'         - initial global step-size, aka mutation strength (`float`).

    Examples
    --------
    Use the optimizer `XNES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy  # engine for numerical computing
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.nes.xnes import XNES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> xnes = XNES(problem, options)  # initialize the optimizer class
       >>> results = xnes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"XNES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       XNES: 5000, 1.3565728021697798e-18

    Attributes
    ----------
    lr_cv         : `float`
                    learning rate of covariance matrix adaptation.
    lr_sigma      : `float`
                    learning rate of global step-size adaptation.
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search/sampling/mutation distribution.
    n_individuals : `int`
                    number of offspring/descendants, aka offspring population size.
    n_parents     : `int`
                    number of parents/ancestors, aka parental population size.
    sigma         : `float`
                    global step-size, aka mutation strength (i.e., overall std of Gaussian search distribution).

    References
    ----------
    Wierstra, D., Schaul, T., Glasmachers, T., Sun, Y., Peters, J. and Schmidhuber, J., 2014.
    Natural evolution strategies.
    Journal of Machine Learning Research, 15(1), pp.949-980.
    https://jmlr.org/papers/v15/wierstra14a.html

    Schaul, T., 2011.
    Studies in continuous black-box optimization.
    Doctoral Dissertation, Technische Universität München.
    https://people.idsia.ch/~schaul/publications/thesis.pdf

    Glasmachers, T., Schaul, T., Yi, S., Wierstra, D. and Schmidhuber, J., 2010, July.
    Exponential natural evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 393-400).
    https://dl.acm.org/doi/abs/10.1145/1830483.1830557

    See the official Python source code from PyBrain:
    https://github.com/pybrain/pybrain/blob/master/pybrain/optimization/distributionbased/xnes.py
    """
    def __init__(self, problem, options):
        options['sigma'] = np.Inf  # not used for `SGES`
        NES.__init__(self, problem, options)
        self.lr_cv = 0.6*(3.0 + np.log(self.ndim_problem))/self.ndim_problem/np.sqrt(self.ndim_problem)
        self.lr_sigma = 0.6*(3.0 + np.log(self.ndim_problem))/self.ndim_problem/np.sqrt(self.ndim_problem)
        self._eye = np.eye(self.ndim_problem)

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        a = np.eye(self.ndim_problem)
        inv_a = np.eye(self.ndim_problem)
        log_det = 0.0
        self._w = np.maximum(0.0, np.log(self.n_individuals/2.0 + 1.0) - np.log(
            self.n_individuals - np.arange(self.n_individuals)))
        return x, y, mean, a, inv_a, log_det

    def iterate(self, x=None, y=None, mean=None, a=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            x[k] = mean + np.dot(a, self.rng_optimization.standard_normal((self.ndim_problem,)))
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, y=None, mean=None, a=None, inv_a=None, log_det=None):
        order = np.argsort(-y)
        u = np.empty((self.n_individuals,))
        for i, o in enumerate(order):
            u[o] = self._w[i]
        u = u/np.sum(u) - 1.0/self.n_individuals
        s = np.empty((self.n_individuals, self.ndim_problem))
        for k in range(self.n_individuals):
            s[k] = np.dot(inv_a, x[k] - mean)
        d_c = np.dot(s.T, u)
        g_cv = np.dot(np.array([np.outer(k, k) - self._eye for k in s]).T, u)
        trace = np.trace(g_cv)
        g_cv -= trace/self.ndim_problem*self._eye
        d_a = 0.5*(self.lr_sigma*trace/self.ndim_problem*self._eye + self.lr_cv*g_cv)
        mean += np.dot(a, d_c)
        a = np.dot(a, scipy.linalg.expm(d_a))
        inv_a = np.dot(scipy.linalg.expm(-d_a), inv_a)
        log_det += 0.5*self.lr_sigma*trace
        return mean, a, inv_a, log_det

    def restart_reinitialize(self, x=None, y=None, mean=None, a=None, inv_a=None, log_det=None):
        if self.is_restart and NES.restart_reinitialize(self, y):
            x, y, mean, a, inv_a, log_det = self.initialize(True)
        return x, y, mean, a, inv_a, log_det

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = NES.optimize(self, fitness_function)
        x, y, mean, a, inv_a, log_det = self.initialize()
        while True:
            x, y = self.iterate(x, y, mean, a, args)
            if self._check_terminations():
                break
            self._print_verbose_info(fitness, y)
            mean, a, inv_a, log_det = self._update_distribution(x, y, mean, a, inv_a, log_det)
            self._n_generations += 1
            x, y, mean, a, inv_a, log_det = self.restart_reinitialize(x, y, mean, a, inv_a, log_det)
        return self._collect(fitness, y, mean)



import numpy as np

from pypop7.optimizers.es.es import ES


class LMMAES(ES):
    """Limited-Memory Matrix Adaptation Evolution Strategy (LMMAES).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'             - initial global step-size, aka mutation strength (`float`),
                * 'mean'              - initial (starting) point, aka mean of Gaussian search distribution
                  (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`).

                * 'n_evolution_paths' - number of evolution paths (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_individuals'     - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'         - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`),
                * 'c_s'               - learning rate of evolution path update (`float`, default:
                  `2.0*options['n_individuals']/problem['ndim_problem']`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.lmmaes import LMMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 200,
       ...            'lower_boundary': -5*numpy.ones((200,)),
       ...            'upper_boundary': 5*numpy.ones((200,))}
       >>> options = {'max_function_evaluations': 500000,  # set optimizer options
       ...            'seed_rng': 0,
       ...            'mean': 3*numpy.ones((200,)),
       ...            'sigma': 0.1,  # the global step-size may need to be tuned for better performance
       ...            'is_restart': False}
       >>> lmmaes = LMMAES(problem, options)  # initialize the optimizer class
       >>> results = lmmaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"LMMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       LMMAES: 500000, 1.0745854362945823e-06

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/4pttkpmc>`_ for more details.

    Attributes
    ----------
    c_s               : `float`
                        learning rate of evolution path update.
    mean              : `array_like`
                        initial (starting) point, aka mean of Gaussian search distribution.
    n_evolution_paths : `int`
                        number of evolution paths.
    n_individuals     : `int`
                        number of offspring, aka offspring population size.
    n_parents         : `int`
                        number of parents, aka parental population size.
    sigma             : `float`
                        final global step-size, aka mutation strength.

    References
    ----------
    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    See the official Python version from Prof. Glasmachers:
    https://www.ini.rub.de/upload/editor/file/1604950981_dc3a4459a4160b48d51e/lmmaes.py
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        n_evolution_paths = 4 + int(3*np.log(self.ndim_problem))
        self.n_evolution_paths = options.get('n_evolution_paths', n_evolution_paths)
        self.c_s = None
        self._s_1 = None
        self._s_2 = None
        self._c_d = 1.0/(self.ndim_problem*np.power(1.5, np.arange(self.n_evolution_paths)))
        self._c_c = None

    def initialize(self, is_restart=False):
        self.c_s = self.options.get('c_s', 2.0*self.n_individuals/self.ndim_problem)
        _s_1 = 1.0 - self.c_s
        if _s_1 < 0.0:  # undefined in the original paper
            _s_1 = 0.5
        self._s_1 = _s_1
        _s_2 = self._mu_eff*self.c_s*(2.0 - self.c_s)
        if _s_2 < 0.0:  # undefined in the original paper
            _s_2 = np.square(0.5)
        self._s_2 = np.sqrt(_s_2)
        self._c_c = self.n_individuals/(self.ndim_problem*np.power(4.0, np.arange(self.n_evolution_paths)))
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        d = np.empty((self.n_individuals, self.ndim_problem))  # search directions
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path
        tm = np.zeros((self.n_evolution_paths, self.ndim_problem))  # transformation matrix M
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._list_initial_mean.append(np.copy(mean))
        return z, d, mean, s, tm, y

    def iterate(self, z=None, d=None, mean=None, tm=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return z, d, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            d[k] = z[k]
            for j in range(np.minimum(self._n_generations, self.n_evolution_paths)):
                d[k] = (1.0 - self._c_d[j])*d[k] + self._c_d[j]*tm[j]*np.dot(tm[j], d[k])
            y[k] = self._evaluate_fitness(mean + self.sigma*d[k], args)
        return z, d, y

    def _update_distribution(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        d_w = np.dot(self._w[:self.n_parents], d[order])
        z_w = np.dot(self._w[:self.n_parents], z[order])
        # update distribution mean
        mean += self.sigma*d_w
        # update evolution path (p_c, s) and low-rank transformation matrix (tm)
        s = self._s_1*s + self._s_2*z_w
        for k in range(self.n_evolution_paths):  # rank-m
            _tm_1 = 1.0 - self._c_c[k]
            if _tm_1 < 0.0:  # undefined in the original paper
                _tm_1 = 0.5
            _tm_2 = self._mu_eff*self._c_c[k]*(2.0 - self._c_c[k])
            if _tm_2 < 0.0:  # undefined in the original paper
                _tm_2 = np.square(0.5)
            tm[k] = _tm_1*tm[k] + np.sqrt(_tm_2)*z_w
        # update global step-size
        self.sigma *= np.exp(self.c_s/2.0*(np.sum(np.square(s))/self.ndim_problem - 1.0))
        return mean, s, tm

    def restart_reinitialize(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            z, d, mean, s, tm, y = self.initialize(True)
        return z, d, mean, s, tm, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, d, mean, s, tm, y = self.initialize()
        while not self.termination_signal:
            # sample and evaluate offspring population
            z, d, y = self.iterate(z, d, mean, tm, y, args)
            if self._check_terminations():
                break
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            mean, s, tm = self._update_distribution(z, d, mean, s, tm, y)
            z, d, mean, s, tm, y = self.restart_reinitialize(z, d, mean, s, tm, y)
        results = self._collect(fitness, y, mean)
        results['s'] = s
        results['tm'] = tm
        return results



import numpy as np  # engine for numerical computing

from pypop7.optimizers.eda.eda import EDA


class AEMNA(EDA):
    """Adaptive Estimation of Multivariate Normal Algorithm (AEMNA).

    .. note:: `AEMNA` learns the *full* covariance matrix of the Gaussian sampling distribution, resulting
       in a *cubic* time complexity w.r.t. each generation. Therefore, like `EMNA`, it is rarely used for
       large-scale black-box optimization (LSBBO). It is **highly recommended** to first attempt other
       more advanced methods for LSBBO.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default: `200`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy  # engine for numerical computing
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.eda.aemna import AEMNA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> aemna = AEMNA(problem, options)  # initialize the optimizer class
       >>> results = aemna.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"AEMNA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       AEMNA: 5000, 0.0023607608362747035

    For its correctness checking of coding, refer to `this code-based repeatability report
    <hhttps://tinyurl.com/5ec2uest>`_ for more details.

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.

    References
    ----------
    Larrañaga, P. and Lozano, J.A. eds., 2002.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5
    """
    def __init__(self, problem, options):
        EDA.__init__(self, problem, options)

    def initialize(self, args=None):
        x = self.rng_optimization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                          size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        order = np.argsort(y)[:self.n_parents]
        mean, cov = np.mean(x[order], axis=0), np.cov(np.transpose(x[order]))
        return x, y, mean, cov

    def iterate(self, x=None, y=None, mean=None, cov=None, args=None):
        xx = self.rng_optimization.multivariate_normal(mean, cov)
        yy = self._evaluate_fitness(xx, args)
        order = np.argsort(y)[:self.n_parents]
        worst = order[-1]
        if yy < y[worst]:
            mean_bak = np.copy(mean)
            mean += (xx - x[worst])/self.n_parents
            ndim2 = np.power(self.n_parents, 2)
            for i in range(self.ndim_problem):
                for j in range(self.ndim_problem):
                    cov[i, j] = (cov[i, j] - ((xx[i] - x[worst, i])*np.sum(x[order, j] - mean_bak[j]))/ndim2 -
                                 ((xx[j] - x[worst, j])*np.sum(x[order, i] - mean_bak[i]))/ndim2 +
                                 ((xx[i] - x[worst, i])*(xx[j] - x[worst, j]))/ndim2 -
                                 ((x[worst, i] - mean[i])*(x[worst, j] - mean[j]))/self.n_parents +
                                 ((xx[i] - mean[i])*(xx[j] - mean[j]))/self.n_parents)
            x[worst], y[worst] = xx, yy
        self._n_generations += 1
        return x, y, mean, cov

    def optimize(self, fitness_function=None, args=None):
        fitness = EDA.optimize(self, fitness_function)
        x, y, mean, cov = self.initialize(args)
        while not self._check_terminations():  # similar to steady-state genetic algorithm
            self._print_verbose_info(fitness, y)
            x, y, mean, cov = self.iterate(x, y, mean, cov, args)
        return self._collect(fitness, y)



from pypop7.optimizers.es.es import ES  # abstract class for all ES classes
from pypop7.optimizers.es.res import RES
from pypop7.optimizers.es.ssaes import SSAES
from pypop7.optimizers.es.dsaes import DSAES
from pypop7.optimizers.es.csaes import CSAES
from pypop7.optimizers.es.saes import SAES
from pypop7.optimizers.es.samaes import SAMAES
from pypop7.optimizers.es.cmaes import CMAES
from pypop7.optimizers.es.opoc2006 import OPOC2006
from pypop7.optimizers.es.sepcmaes import SEPCMAES
from pypop7.optimizers.es.opoc2009 import OPOC2009
from pypop7.optimizers.es.ccmaes2009 import CCMAES2009
from pypop7.optimizers.es.opoa2010 import OPOA2010
from pypop7.optimizers.es.vdcma import VDCMA
from pypop7.optimizers.es.lmcmaes import LMCMAES
from pypop7.optimizers.es.opoa2015 import OPOA2015
from pypop7.optimizers.es.ccmaes2016 import CCMAES2016
from pypop7.optimizers.es.vkdcma import VKDCMA
from pypop7.optimizers.es.lmcma import LMCMA
from pypop7.optimizers.es.maes import MAES
from pypop7.optimizers.es.r1es import R1ES
from pypop7.optimizers.es.rmes import RMES
from pypop7.optimizers.es.lmmaes import LMMAES
from pypop7.optimizers.es.fmaes import FMAES
from pypop7.optimizers.es.fcmaes import FCMAES
from pypop7.optimizers.es.ddcma import DDCMA
from pypop7.optimizers.es.mmes import MMES


__all__ = [ES,  # Evolution Strategies
           RES,  # Rechenberg’s (1+1)-Evolution Strategy with 1/5th success rule
           SSAES,  # Schwefel's Self-Adaptation Evolution Strategy
           DSAES,  # Derandomized Self-Adaptation Evolution Strategy
           CSAES,  # Cumulative Step-size self-Adaptation Evolution Strategy
           SAES,  # Self-Adaptation Evolution Strategy
           SAMAES,  # Self-Adaptation Matrix Adaptation Evolution Strategy
           CMAES,  # Covariance Matrix Adaptation Evolution Strategy
           OPOC2006,  # (1+1)-Cholesky-CMA-ES (2006)
           OPOC2009,  # (1+1)-Cholesky-CMA-ES (2009)
           CCMAES2009,  # Cholesky-CMA-ES (2009)
           OPOA2010,  # (1+1)-Active-CMA-ES (2010)
           OPOA2015,  # (1+1)-Active-CMA-ES (2015)
           CCMAES2016,  # Cholesky-CMA-ES (2016)
           MAES,  # Matrix Adaptation Evolution Strategy
           FMAES,  # Fast Matrix Adaptation Evolution Strategy
           DDCMA,  # Diagonal Decoding Covariance Matrix Adaptation
           SEPCMAES,  # Separable Covariance Matrix Adaptation Evolution Strategy
           VDCMA,  # Linear Covariance Matrix Adaptation
           LMCMAES,  # Limited-Memory CMA-ES
           VKDCMA,  # Restricted Covariance Matrix Adaptation
           LMCMA,  # Limited-Memory Covariance Matrix Adaptation
           R1ES,  # Rank-One Evolution Strategy
           RMES,  # Rank-M Evolution Strategy
           LMMAES,  # Limited-Memory Matrix Adaptation Evolution Strategy
           FCMAES,  # Fast Covariance Matrix Adaptation Evolution Strategy
           MMES]  # Mixture Model-based Evolution Strategy



import numpy as np
from scipy.stats import cauchy

from pypop7.optimizers.de.de import DE


class JADE(DE):
    """Adaptive Differential Evolution (JADE).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default: `100`),
                * 'mu'            - mean of normal distribution for adaptation of crossover probability (`float`,
                  default: `0.5`),
                * 'median'        - median of Cauchy distribution for adaptation of mutation factor (`float`,
                  default: `0.5`),
                * 'p'             - level of greediness of mutation strategy (`float`, default: `0.05`),
                * 'c'             - life span (`float`, default: `0.1`),
                * 'is_bound'      - flag to limit all samplings inside the search range (`boolean`, default: `False`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.de.jade import JADE
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 0}
       >>> jade = JADE(problem, options)  # initialize the optimizer class
       >>> results = jade.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"JADE: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       JADE: 5000, 4.844728910084905e-05

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/bddsba5s>`_ for more details.

    Attributes
    ----------
    c             : `float`
                    life span.
    is_bound      : `boolean`
                    flag to limit all samplings inside the search range.
    median        : `float`
                    median of Cauchy distribution for adaptation of mutation factor.
    mu            : `float`
                    mean of normal distribution for adaptation of crossover probability.
    n_individuals : `int`
                    number of offspring, offspring population size.
    p             : `float`
                    level of greediness of mutation strategy.

    References
    ----------
    Zhang, J., and Sanderson, A. C. 2009.
    JADE: Adaptive differential evolution with optional external archive.
    IEEE Transactions on Evolutionary Computation, 13(5), pp.945–958.
    https://ieeexplore.ieee.org/document/5208221/
    """
    def __init__(self, problem, options):
        DE.__init__(self, problem, options)
        self.mu = options.get('mu', 0.5)  # mean of normal distribution for adaptation of crossover probabilities
        self.median = options.get('median', 0.5)  # location of Cauchy distribution for adaptation of mutation factor
        self.p = options.get('p', 0.05)  # level of greediness of the mutation strategy
        assert 0.0 <= self.p <= 1.0
        self.c = options.get('c', 0.1)  # life span
        assert 0.0 <= self.c <= 1.0
        self.is_bound = options.get('is_bound', False)

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        a = np.empty((0, self.ndim_problem))  # set of archived inferior solutions
        return x, y, a

    def bound(self, x=None, xx=None):
        if not self.is_bound:
            return x
        for k in range(self.n_individuals):
            idx = np.array(x[k] < self.lower_boundary)
            if idx.any():
                x[k][idx] = (self.lower_boundary + xx[k])[idx]/2.0
            idx = np.array(x[k] > self.upper_boundary)
            if idx.any():
                x[k][idx] = (self.upper_boundary + xx[k])[idx]/2.0
        return x

    def mutate(self, x=None, y=None, a=None):
        x_mu = np.empty((self.n_individuals,  self.ndim_problem))  # mutated population
        f_mu = np.empty((self.n_individuals,))  # mutated mutation factors
        order = np.argsort(y)[:int(np.ceil(self.p*self.n_individuals))]  # index of the [100*p]% best individuals
        x_p = x[self.rng_optimization.choice(order, (self.n_individuals,))]
        x_un = np.vstack((np.copy(x), a))  # archive
        for k in range(self.n_individuals):
            f_mu[k] = cauchy.rvs(loc=self.median, scale=0.1, random_state=self.rng_optimization)
            while f_mu[k] <= 0.0:
                f_mu[k] = cauchy.rvs(loc=self.median, scale=0.1, random_state=self.rng_optimization)
            if f_mu[k] > 1.0:
                f_mu[k] = 1.0
            r1 = self.rng_optimization.choice([i for i in range(self.n_individuals) if i != k])
            r2 = self.rng_optimization.choice([i for i in range(len(x_un)) if i != k and i != r1])
            x_mu[k] = x[k] + f_mu[k]*(x_p[k] - x[k]) + f_mu[k]*(x[r1] - x_un[r2])
        return x_mu, f_mu

    def crossover(self, x_mu=None, x=None):
        x_cr = np.copy(x)
        p_cr = self.rng_optimization.normal(self.mu, 0.1, (self.n_individuals,))  # crossover probabilities
        # truncate to [0, 1]
        p_cr = np.minimum(np.maximum(p_cr, 0.0), 1.0)
        for k in range(self.n_individuals):
            i_rand = self.rng_optimization.integers(self.ndim_problem)
            for i in range(self.ndim_problem):
                if (i == i_rand) or (self.rng_optimization.random() < p_cr[k]):
                    x_cr[k, i] = x_mu[k, i]
        return x_cr, p_cr

    def select(self, args=None, x=None, y=None, x_cr=None, a=None, f_mu=None, p_cr=None):
        f = np.empty((0,))  # set of all successful mutation factors
        p = np.empty((0,))  # set of all successful crossover probabilities
        for k in range(self.n_individuals):
            if self._check_terminations():
                break
            yy = self._evaluate_fitness(x_cr[k], args)
            if yy < y[k]:
                a = np.vstack((a, x[k]))  # archive of the inferior solution
                f = np.hstack((f, f_mu[k]))  # archive of the successful mutation factor
                p = np.hstack((p, p_cr[k]))  # archive of the successful crossover probability
                x[k] = x_cr[k]
                y[k] = yy
        if len(p) != 0:  # for mean update of normal distribution
            self.mu = (1.0 - self.c)*self.mu + self.c*np.mean(p)
        if len(f) != 0:  # for location update of Cauchy distribution
            self.median = (1.0 - self.c)*self.median + self.c*np.sum(np.power(f, 2))/np.sum(f)
        return x, y, a

    def iterate(self, x=None, y=None, a=None, args=None):
        x_mu, f_mu = self.mutate(x, y, a)
        x_cr, p_cr = self.crossover(x_mu, x)
        x_cr = self.bound(x_cr, x)
        x, y, a = self.select(args, x, y, x_cr, a, f_mu, p_cr)
        # randomly remove solutions to keep the archive size fixed
        if len(a) > self.n_individuals:
            a = np.delete(a, self.rng_optimization.choice(len(a), (len(a) - self.n_individuals,), False), 0)
        self._n_generations += 1
        return x, y, a

    def optimize(self, fitness_function=None, args=None):
        fitness = DE.optimize(self, fitness_function)
        x, y, a = self.initialize(args)
        while not self._check_terminations():
            self._print_verbose_info(fitness, y)
            x, y, a = self.iterate(x, y, a, args)
        return self._collect(fitness, y)



"""Repeat the following paper for `SCEM`:
    Kroese, D.P., Porotsky, S. and Rubinstein, R.Y., 2006.
    The cross-entropy method for continuous multi-extremal optimization.
    Methodology and Computing in Applied Probability, 8(3), pp.383-407.
    https://link.springer.com/article/10.1007/s11009-006-9753-0
    (See [Appendix B Main CE Program] for the official Matlab code.)

    Luckily our Python code could repeat the data generated by the original Matlab code *well*.
    Therefore, we argue that its repeatability could be **well-documented**.



    You can run the following Matlab script (note that save function `Rosen` as a separate file):
    ---------------------------------------------------------------------------------------------
    function out = Rosen(X)
    r=[];
    for i = 1:size(X,2)-1
    r = [100*(X(:,i+1)-X(:,i).^2).^2+(X(:,i)-1).^2,r];
    end
    out = sum(r,2);
    end

    n=1000;
    N = 1000;
    Nel = 200;
    alpha = 0.8;
    mu = 2*ones(1,n);
    sigma = 10*ones(1,n);
    mu_last = mu;
    sigma_last = sigma;
    S_best_overall = Inf;
    t = 0;
    while t < 1001
    t = t + 1;
    mu = alpha*mu + (1-alpha)*mu_last;
    sigma= alpha*sigma + (1-alpha)*sigma_last;
    X = ones(N,1)*mu + randn(N,n)*diag(sigma);
    SA = Rosen(X);
    [S_sort,I_sort] = sort(SA);
    S_best = S_sort(1);
    if (S_best < S_best_overall)
    S_best_overall = S_best;
    end
    mu_last = mu;
    sigma_last = sigma;
    Xel = X(I_sort(1:Nel),:);
    mu = mean(Xel);
    sigma = std(Xel);
    end
    fprintf('%9.8f\n', S_best_overall);
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock
from pypop7.optimizers.cem.scem import SCEM


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 1000
    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5*np.ones((ndim_problem,)),
               'upper_boundary': 5*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1000*1000,
               'mean': 2.0*np.ones((ndim_problem,)),
               'seed_rng': 0,
               'sigma': 10.0,
               'verbose': 200,
               'saving_fitness': 50000}
    scem = SCEM(problem, options)
    results = scem.optimize()
    print(results)  # 1044.879869712122 vs 1050.51959312 (from the Matlab code)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



import numpy as np

from pypop7.optimizers.rs.rs import RS


class PRS(RS):
    """Pure Random Search (PRS).

    .. note:: `PRS` is one of the *simplest* and *earliest* black-box optimizers, dating back to at least
       `1950s <https://pubsonline.informs.org/doi/abs/10.1287/opre.6.2.244>`_. Although recently it
       has been successfully applied in several *relatively low-dimensional* problems (particularly
       `hyper-parameter optimization <https://www.jmlr.org/papers/v13/bergstra12a.html>`_), it generally
       suffers from the famous **curse of dimensionality** for large-scale black-box optimization (LSBBO),
       owing to the lack of *adaptation*, a highly desirable property for most sophisticated search
       algorithms. Therefore, it is **highly recommended** to first attempt more advanced (e.g.
       population-based) methods for LSBBO.

       Here we include it mainly for *benchmarking* purpose. As pointed out in `Probabilistic Machine Learning
       <https://probml.github.io/pml-book/book2.html>`_, *"this should always be tried as a baseline"*.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`).

    Examples
    --------
    Use the `PRS` optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.rs.prs import PRS
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> prs = PRS(problem, options)  # initialize the optimizer class
       >>> results = prs.optimize()  # run the optimization process
       >>> # return the number of used function evaluations and found best-so-far fitness
       >>> print(f"PRS: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       PRS: 5000, 0.11497678820610932

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/mrx2kffy>`_ for more details.

    References
    ----------
    Bergstra, J. and Bengio, Y., 2012.
    Random search for hyper-parameter optimization.
    Journal of Machine Learning Research, 13(2).
    https://www.jmlr.org/papers/v13/bergstra12a.html

    Schmidhuber, J., Hochreiter, S. and Bengio, Y., 2001.
    Evaluating benchmark problems by random guessing.
    A Field Guide to Dynamical Recurrent Networks, pp.231-235.
    https://ml.jku.at/publications/older/ch9.pdf

    Brooks, S.H., 1958.
    A discussion of random methods for seeking maxima.
    Operations Research, 6(2), pp.244-251.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.6.2.244
    """
    def __init__(self, problem, options):
        RS.__init__(self, problem, options)
        # set default: 1 -> uniformly distributed random sampling
        self._sampling_type = options.get('_sampling_type', 1)
        if self._sampling_type not in [0, 1]:  # 0 -> normally distributed random sampling
            info = 'For currently {:s}, only support uniformly or normally distributed random sampling.'
            raise ValueError(info.format(self.__class__.__name__))
        elif self._sampling_type == 0:
            self.sigma = options.get('sigma')  # initial global step-size (fixed during optimization)
            assert self.sigma is not None

    def _sample(self, rng):
        if self._sampling_type == 0:
            x = self.x + self.sigma*rng.standard_normal(size=(self.ndim_problem,))
        else:
            x = rng.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        return x

    def initialize(self):
        if self.x is None:
            x = self._sample(self.rng_initialization)
        else:
            x = np.copy(self.x)
        assert len(x) == self.ndim_problem
        return x

    def iterate(self):  # individual-based sampling
        return self._sample(self.rng_optimization)



"""Repeat the following paper for `BES`:
    Gao, K. and Sener, O., 2022, June.
    Generalizing Gaussian Smoothing for Random Search.
    In International Conference on Machine Learning (pp. 7077-7101). PMLR.
    https://proceedings.mlr.press/v162/gao22f.html
    https://icml.cc/media/icml-2022/Slides/16434.pdf

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/bes

    Since the current Python implementation is only a *simplified* version of the original BES algorithm
    without noisy function (fitness) evaluations, its repeatability **cannot** be guaranteed.

    However, we found that the current implementation could show *very similar* performance on the same
    benchmark functions (after removing its noise part). The resulting much less (>100x) number of function
    evaluations needed may be well explained by the much easier fitness landscape (i.e., without noisiness).
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere, rosenbrock
from pypop7.optimizers.rs.bes import BES


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')

    problem = {'fitness_function': sphere,
               'ndim_problem': 100}
    options = {'max_function_evaluations': 100*101,
               'seed_rng': 0,  # undefined in the original paper
               'x': np.random.default_rng(1).standard_normal(size=100),
               'n_individuals': 10,
               'saving_fitness': 101}
    bes = BES(problem, options)
    results = bes.optimize()
    print(results)
    plt.figure()
    plt.xlabel('Rounds')
    plt.xticks([0, 20, 40, 60, 80, 100])
    plt.ylabel('Oracle')
    plt.yticks([0, 20, 40, 60, 80, 100])
    plt.plot(results['fitness'][:, 0]/101, results['fitness'][:, 1], 'k')
    plt.show()

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': 100,
               'lower_boundary': -2*np.ones((100,)),
               'upper_boundary': 2*np.ones((100,))}
    options = {'max_function_evaluations': 100*101,
               'seed_rng': 2,  # undefined in the original paper
               'x': np.random.default_rng(3).standard_normal(size=100),
               'n_individuals': 10,
               'c': 0.1,
               'lr': 0.000001,
               'saving_fitness': 101}
    bes = BES(problem, options)
    results = bes.optimize()
    print(results)
    plt.figure()
    plt.xlabel('Rounds')
    plt.xticks([0, 20, 40, 60, 80, 100])
    plt.ylabel('Oracle')
    plt.yticks([10000, 20000, 30000, 40000])
    plt.plot(results['fitness'][:, 0]/101, results['fitness'][:, 1], 'k')
    plt.show()



import numpy as np

from pypop7.optimizers.es.es import ES


class RES(ES):
    """Rechenberg's (1+1)-Evolution Strategy with 1/5th success rule (RES).

    .. note:: `RES` is the *first* ES with self-adaptation of the *global* step-size, designed by Rechenberg, one
       recipient of `IEEE Evolutionary Computation Pioneer Award 2002 <https://tinyurl.com/456as566>`_. As
       **theoretically** investigated in Rechenberg's seminal PhD dissertation, the existence of narrow **evolution
       window** explains the necessarity of step-size *adaptation* to maximize local convergence progress, if possible.

       Since there is only one parent and only one offspring for each generation, `RES` generally shows very
       limited *exploration* ability for large-scale black-box optimization (LSBBO). Therefore, it is **highly
       recommended** to first attempt more advanced ES variants (e.g. `LMCMA`, `LMMAES`) for LSBBO. Here we
       include it mainly for *benchmarking* and *theoretical* purpose.

       AKA two-membered ES.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'    - initial global step-size, aka mutation strength (`float`),
                * 'mean'     - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'lr_sigma' - learning rate of global step-size self-adaptation (`float`, default:
                  `1.0/np.sqrt(problem['ndim_problem'] + 1.0)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.res import RES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> res = RES(problem, options)  # initialize the optimizer class
       >>> results = res.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"RES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       RES: 5000, 0.06701744137207027

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/5n6ndrn7>`_ for more details.

    Attributes
    ----------
    lr_sigma : `float`
               learning rate of global step-size self-adaptation.
    mean     : `array_like`
               initial (starting) point, aka mean of Gaussian search distribution.
    sigma    : `float`
               final global step-size, aka mutation strength.

    References
    ----------
    Auger, A., Hansen, N., López-Ibáñez, M. and Rudolph, G., 2022.
    Tributes to Ingo Rechenberg (1934--2021).
    ACM SIGEVOlution, 14(4), pp.1-4.
    https://dl.acm.org/doi/10.1145/3511282.3511283
    
    Agapie, A., Solomon, O. and Giuclea, M., 2021.
    Theory of (1+1) ES on the RIDGE.
    IEEE Transactions on Evolutionary Computation, 26(3), pp.501-511.
    https://ieeexplore.ieee.org/abstract/document/9531957

    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44

    Kern, S., Müller, S.D., Hansen, N., Büche, D., Ocenasek, J. and Koumoutsakos, P., 2004.
    Learning probability distributions in continuous evolutionary algorithms–a comparative review.
    Natural Computing, 3, pp.77-112.
    https://link.springer.com/article/10.1023/B:NACO.0000023416.59689.4e

    Beyer, H.G. and Schwefel, H.P., 2002.
    Evolution strategies–A comprehensive introduction.
    Natural Computing, 1(1), pp.3-52.
    https://link.springer.com/article/10.1023/A:1015059928466

    Rechenberg, I., 2000.
    Case studies in evolutionary experimentation and computation.
    Computer Methods in Applied Mechanics and Engineering, 186(2-4), pp.125-140.
    https://www.sciencedirect.com/science/article/pii/S0045782599003813

    Rechenberg, I., 1989.
    Evolution strategy: Nature’s way of optimization.
    In Optimization: Methods and Applications, Possibilities and Limitations (pp. 106-126).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-83814-9_6

    Rechenberg, I., 1984.
    The evolution strategy. A mathematical model of darwinian evolution.
    In Synergetics—from Microscopic to Macroscopic Order (pp. 122-132). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-69540-7_13
    """
    def __init__(self, problem, options):
        options['n_parents'] = 1  # mandatory setting
        options['n_individuals'] = 1  # mandatory setting
        ES.__init__(self, problem, options)
        if self.lr_sigma is None:
            self.lr_sigma = 1.0/np.sqrt(self.ndim_problem + 1.0)
        assert self.lr_sigma > 0, f'`self.lr_sigma` = {self.lr_sigma}, but should > 0.'

    def initialize(self, args=None, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        y = self._evaluate_fitness(mean, args)  # fitness
        best_so_far_y = np.copy(y)
        self._list_initial_mean.append(np.copy(mean))
        return mean, y, best_so_far_y

    def iterate(self, args=None, mean=None):  # to sample and evaluate only one offspring
        x = mean + self.sigma*self.rng_optimization.standard_normal((self.ndim_problem,))
        y = self._evaluate_fitness(x, args)
        return x, y

    def restart_reinitialize(self, args=None, mean=None, y=None, best_so_far_y=None, fitness=None):
        if not self.is_restart:
            return mean, y, best_so_far_y
        self._list_fitness.append(best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._list_fitness) >= self.stagnation:
            is_restart_2 = (self._list_fitness[-self.stagnation] - self._list_fitness[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._print_verbose_info(fitness, y, True)
            if self.verbose:
                print(' ....... *** restart *** .......')
            self._n_restart += 1
            self._list_generations.append(self._n_generations)  # for each restart
            self._n_generations = 0
            self.sigma = np.copy(self._sigma_bak)
            mean, y, best_so_far_y = self.initialize(args, True)
            self._list_fitness = [best_so_far_y]
        return mean, y, best_so_far_y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, y, best_so_far_y = self.initialize(args)
        while not self.termination_signal:
            self._print_verbose_info(fitness, y)
            x, y = self.iterate(args, mean)
            self._n_generations += 1
            if self._check_terminations():
                break
            self.sigma *= np.power(np.exp(float(y < best_so_far_y) - 0.2), self.lr_sigma)
            if y <= best_so_far_y:
                mean, best_so_far_y = x, y
            mean, y, best_so_far_y = self.restart_reinitialize(args, mean, y, best_so_far_y, fitness)
        return self._collect(fitness, y, mean)



"""Repeat the following paper for `CPSO`:
    Van den Bergh, F. and Engelbrecht, A.P., 2004.
    A cooperative approach to particle swarm optimization.
    IEEE Transactions on Evolutionary Computation, 8(3), pp.225-239.
    https://ieeexplore.ieee.org/document/1304845

    Luckily our Python code could repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, griewank
from pypop7.optimizers.pso.cpso import CPSO as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 30

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -2.048*np.ones((ndim_problem,)),
               'upper_boundary': 2.048*np.ones((ndim_problem,))}
    options = {'seed_rng': 0,  # not given in the original paper
               'max_function_evaluations': 2e5,
               'fitness_threshold': 0.0}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 2.11884e-1 vs 9.06e-1 (from original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'seed_rng': 0,  # not given in the original paper
               'max_function_evaluations': 2e5,
               'fitness_threshold': 0.0}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 1.23742e-2 vs 2.25e-2 (from original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))
